// @flow
import Stats from 'stats.js';
import SpikeWarsArtist from './spike-wars-artist';
import {update, render} from './game-logic';

import type Board from './game-types';
import type GameState from './game-types';
import type Piece from './game-types';

const WALL = "WALL";
const DEBUG = "DEBUG";

type SpikeWarsOptions = {fps?: number, stats?: Object};

export default class SpikeWars {
  rows: number;
  cols: number;

  // This is where we keep track of all the things that are happening in the
  // game.
  state: GameState;

  nextGameTick: number; // At what time does the next game update happen?
  skipTicks: number;    // How long does the game wait between updates?

  // This is the thing that knows how to draw stuff.
  artist: SpikeWarsArtist;

  stats: ?Object;
  pauseCheckbox: HTMLInputElement;
  paused: boolean;

  // The constructor is the function that is run when somebody first creates a
  // new SpikeWars instance. This is where we set stuff up for the first time.
  constructor({fps = 30, stats}: SpikeWarsOptions) {

    // Get our artist
    this.rows = 16;
    this.cols = 40;
    this.artist = new SpikeWarsArtist({rows: this.rows, columns: this.cols});

    this.state = {board: this.newBoard(), count: 0}; // create a new board

    this.stats = stats;

    this.pauseCheckbox = window.document.getElementsByTagName('input')[0];
    this.pauseCheckbox.checked = false;
    this.pauseCheckbox.addEventListener('change', () => { this.toggleGamePaused() });
    this.paused = false;

    this.skipTicks = Math.ceil(1000 / fps);
    // TODO: register event handlers on the canvas element.
  }

  // This function creates a new board.
  newBoard(): Board {
    var board = new Array(this.rows * this.cols);
    for (var i = 0; i < board.length; i++) {
      var col = i % this.cols;
      var row = Math.floor(i / this.cols);
      if (col === 0 || col === (this.cols - 1) ||
          row === 0 || row === (this.rows - 1)) board[i] = WALL; }
      return board; }

  toggleGamePaused(): void {
    this.paused = this.pauseCheckbox.checked;

    if (!this.paused) {
      this.nextGameTick = new Date().getTime();
      this.gameLoop(); }
  }

  // The game loop is where the game actually runs. It is called a loop because
  // it runs over and over and over, in a loop until it is time to stop the
  // game.
  gameLoop(): void {
    if (this.paused) return;

    // We are going to ask the browser to call us again next time we have a
    // chance to draw. We're doing this first because we want to make sure we
    // reserve our place in line before we get busy doing stuff.
    window.requestAnimationFrame(() => { this.gameLoop(); });
    //window.setTimeout(() => { this.gameLoop(); });

    // Now we need to check if it is time to do stuff. If our computer is
    // really fast the game loop might run really fast, or if our computer is
    // busy the speed might change. That messes up the game (we want it to just
    // go one speed). So to fix this we use a kind of clock. We figure out how
    // fast we want the game to run. Then we figure out how much time we should
    // wait in between each game "tick" (like a clock).
    // 
    // Is it time for the next game tick?
    if ((new Date()).getTime() > this.nextGameTick) {
      if (this.stats) this.stats.begin();

      // Yes! So the first thing we need to do is figure out when the next tick
      // should be.
      this.nextGameTick += this.skipTicks; 

      // Then we update our game state.
      this.state = update(this.state);

      // And we re-draw the world.
      render(this.state, this.artist);
      if (this.stats) this.stats.end(); }
  }

  startGame(): void {
    this.nextGameTick = new Date().getTime();
    this.drawBoard(this.artist, this.state.board);
    this.gameLoop(); }

  drawBoard(artist: SpikeWarsArtist, board: Board): void {
    for (var i = 0; i < board.length; i++) {
      var color: string;
      switch (board[i]) {
        case WALL: color = 'black'; break;
        case DEBUG: color = 'red'; break;
        default: color = 'white';
      }
      this.artist.drawSolidColorIdx(i, color); } }
}


// instantiate the instance of SpikeWars
if (false) {
  var stats = new Stats();
  stats.showPanel(0);
  document.body.appendChild(stats.dom);
  window.spikeWars = new SpikeWars({fps: 30, stats: stats}); }

else { window.spikeWars = new SpikeWars({fps: 30}); }

//window.spikeWars.startGame();

